/**
  Finite Elements with OpenGL/GLUT
  
  Dr. Orion Sky Lawlor, olawlor@acm.org, 2009-03-25 (Public Domain)
*/
#include "GL/glew.h" /* GL extension wrangler: http://glew.sourceforge.net/ */
#include "GL/glut.h" /* windows, mice, etc.  http://freeglut.sourceforge.net */
#include <stdlib.h> /* for "rand" */
#include <stdio.h> /* for printf */
#include "osl/vec4.h" /* 4D & 3D vector class */
#include "osl/mat2.h" /* 2D matrix and vector class */
using osl::mat2;
vec3 make3(const vec2 &v) {return vec3(v.x,v.y,0.0);}

#include "ogl/glsl.h" /* programmable shaders */
#include "ogl/glsl.cpp"

#include "ogl/minicam.h" /* small camera control functions */

#include <vector> /* std::vector container */

#ifndef M_PI
#define M_PI 3.14159265358979323
#endif

/* Random Float: Return a random number between 0 and 1 */
float rf(void) { return (rand()&0xfff)*(1.0/0xfff); }
/* Random Vector: return a random vector in 3D, all axes between 0 and 1 */
vec3 rv(void) {return vec3(rf(),rf(),rf());}

/* Return the wall clock time, in seconds */
double oglTimer(void) {return 0.001*glutGet(GLUT_ELAPSED_TIME);}
/* Check if an error has happened--if so, exit */
void oglCheck(const char *where) {
	int err=0;
	if (0!=(err=glGetError())) {
		fprintf(stderr,"OpenGL error %d at %s\n",err,where);
		exit(1);
	}
}

/* 2x2 matrix eigendecomposition: computes
	A = Q D Q^-1
  where D is a diagonal matrix, and Q is a rotation matrix.
*/
void eigen(const mat2 &A,mat2 &Q,mat2 &D)
{
	double a=A.x.x, c=A.y.x;
	double b=A.x.y, d=A.y.y;

/* Quadratic equation, to solve for eigen values V and vectors v:
	A v = V v
	(A - V I ) v = 0
	det(A - V I ) = 0
	(a-V)*(d-V)-c*d = 0 
	V^2 - (a+d)*V + (a*d-c*b);
*/
	double b__2=(a+d)*0.5; 
	double detSqr=b__2*b__2-(a*d-c*b);
	float det;
	if (detSqr<0) { /* negative determinant */
		//printf("Error: matrix eigenvalues are complex.\n");
		//exit(1);
		// should be Complex(0,sqrt(-detSqr));
		D=mat2(1,0,0,1);
		Q=mat2(1.0);
		return;
	}
	else {
		det=sqrt(detSqr);
	}

	float vals[2]; // eigenvalues
	vals[0]=b__2+det;
	vals[1]=b__2-det;
	D=mat2(vals[0],0,  0,vals[1]);

	for (int eig=0;eig<2;eig++) {
		vec2 v; /* eigenvector */
		if (a!=0 && c!=0) { // Eigenvector uniquely determined by first row:
        		v.x=-c;
        		v.y=a-vals[eig];
		} else if (b!=0 && d!=0) { // Uniquely determined by second row:
        		v.x=d-vals[eig];
        		v.y=-b;
		} else { // Eigenvectors not uniquely determined-- pick one
        		v.x=!eig;
        		v.y=eig;
		}
		float l=length(v);
		Q[eig]=v/l;
	}
}

class particle {
public:
	vec3 P; /* position */
	vec3 V; /* velocity */
	vec2 F; /* temporary summed-up force from all my triangles */
	float mass; /* 1/3 * area of surrounding triangles * density */
	vec3 C; /* color */
};

/* One finite-element triangle.  See Fedkiw et al 2004 paper for the basic idea. */
class triangle {
public:
	int corner[3]; /* array indexes of my particles */
	
	// Look up edge i: from corner i to corner i+1 
	vec2 edge(const particle *pts,int i) {
		return vec2(pts[corner[(i+1)%3]].P - pts[corner[i]].P);
	}
	
	// Normal of edge i, times the length of edge i
	vec2 edgenormal(const particle *pts,int i) {
		vec2 e=edge(pts,i);
		return vec2(-e.y,e.x);
	}
	
	// Body matrix: input barycentric triangle coords; output world coords
	mat2 body_matrix(const particle *pts) {
		return mat2(edge(pts,0),edge(pts,1));
	}
	
	mat2 rest_inverse; /* inverse of body matrix at startup */
	mat2 strain; /* strain matrix: how I've been squished */
	mat2 stress; /* stress matrix: how I push back (lbs/inch) */
	mat2 rotate; /* our element's rotation to get to world coordinates */
	float mass; /* mass of triangle */
	triangle() :stress(0.0) {}
	
	// Compute initial strain ("rest" configuration)
	void setup(particle *pts) {  
		mat2 rest=body_matrix(pts);
		rest_inverse=inverse(rest);
		vec2 a=edge(pts,0), b=edge(pts,1);
		float density=1000.0; /* mass per unit area */
		mass=0.5*fabs(a.x*b.y-a.y*b.x)*density;
		for (int c=0;c<3;c++) /* account mass of triangle to its nodes */
			pts[corner[c]].mass+=(1.0/3.0)*mass;
	}
	
	// Compute our stress, and apply forces to these particles
	void physics(particle *pts) {
		strain=body_matrix(pts)*rest_inverse;
		
		// Rotate off the bulk strain rotation:
		//  Try to make strain's X axis be aligned with the world X (ditto Y)
		vec2 rx=normalize(strain.x), ry=normalize(strain.y);
		rotate=0.5*( // average X and Y rotation matrix
			mat2(rx.x,-rx.y,  rx.y,rx.x)+ // align X
			mat2(ry.y,ry.x,  -ry.x,ry.y) // align Y 
		); 
		strain=strain*rotate; // strain is now expressed in world coordinates
		
		// Material constitutive model: see
		//    http://www.efunda.com/formulae/solid_mechanics/mat_mechanics/hooke_plane_stress.cfm
		float x_squish=strain.x.x-1.0;
		float y_squish=strain.y.y-1.0;
		float shear=strain.y.x+strain.x.y;
		shear*=0.25; /* shear strength is lower in real materials */
		
		float stiffness=1000.0;
		stress=stiffness*(
			mat2(x_squish,shear,  shear,y_squish)
		);
		
		// Apply the stress along each edge to get edge forces
		vec2 e[3],c[3]; // forces along each edge, corner
		e[0]=stress*edgenormal(pts,0);
		e[1]=stress*edgenormal(pts,1);
		e[2]=stress*edgenormal(pts,2);
		
		c[0]=e[2]+e[0];
		c[1]=e[0]+e[1];
		c[2]=e[1]+e[2];
		
		// Apply dissipative force to damp out velocity differences (for stability)
		for (int e=0;e<3;e++) {
			vec2 closing=vec2(pts[corner[(e+1)%3]].V-pts[corner[e]].V);
			vec2 cf=-8.0*closing; 
			c[(e+1)%3]+=cf;
			c[e]-=cf;
		}
		
		// Write back corner forces to particles
		pts[corner[0]].F+=c[0];
		pts[corner[1]].F+=c[1];
		pts[corner[2]].F+=c[2];
	}
	
	// Show our stress and strain configuration onscreen
	void draw(const particle *pts)
	{
		glBegin(GL_LINES);
		vec2 center=vec2((1.0/3)*(
			pts[corner[0]].P+
			pts[corner[1]].P+
			pts[corner[2]].P
		));
	
		for (float ang=0;ang<2*M_PI-0.001;ang+=M_PI*0.5) 
		{
			float angdark=0.3+ang*0.1;
			vec2 dir(cos(ang),sin(ang));
			
			glColor3f(0,1*angdark,0); /* strain in green */
			glVertex2fv(center);
			glVertex2fv(center+strain*dir*0.1);
			
			glColor3f(1*angdark,0,0); /* stress in red */
			glVertex2fv(center);
			glVertex2fv(center+stress*dir*0.01);
			
			glColor3f(1*angdark,1*angdark,0); /* rotation in yellow */
			glVertex2fv(center);
			glVertex2fv(center+rotate*dir*0.1);
		}
		
	
		if (false)
		for (int e=0;e<3;e++) {
			glColor3f(0,0,1); /* blue edge vectors */
			vec2 start=0.5*vec2(pts[corner[e]].P+pts[corner[(e+1)%3]].P);
			glVertex2fv(start);
			glVertex2fv(start+0.1*edgenormal(pts,e));

			glColor3f(1,1,1); /* white edge forces */
			glVertex2fv(start);
			glVertex2fv(start+0.1*stress*edgenormal(pts,e));
			
			start=vec2(pts[corner[e]].P); /* white vertex forces */
			glVertex2fv(start);
			glVertex2fv(start+0.1*stress*(
				edgenormal(pts,e)+edgenormal(pts,(e+3-1)%3)
			));
		}
		
		glEnd();
	}
};



void display(void) /* Draw everything in our window */
{
	unsigned int i;
	oglCheck("before display function");
	static double elapsed=0.0; /* last elapsed display time (seconds) */
	static float dt=0.001, display_dt=0.0;
	static float sim_time=0.0;
	float big=30; /* size of the ground */
	float floor=0.0; /* Z height of floor */
	glFinish(); double start_t=oglTimer();
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA,GL_ONE_MINUS_SRC_ALPHA);
	
/* Particle state storage */
	static bool first_time=true; /* for initial bc's! */
	static std::vector<particle> pts;
	static std::vector<triangle> triangles;
	if (first_time) { /* First time: initialize points */
		static int count=0;
		srand(count++);
		std::ifstream nf("mesh/image.1.node");  std::string crap;
		
		if (!nf) { /* hardcoded mesh: one triangle */
			for (int i=0;i<3;i++) {
				particle p;
				p.P=0.6*rv(); 
				p.P.z=0.0001;
				p.V=0.03*(rv()-vec3(0.5)); 
				p.mass=1.0;
				p.C=rv();
				pts.push_back(p);

				if (i-2>=0) {
					triangle t;
					t.corner[0]=i-2;
					t.corner[1]=i-1;
					t.corner[2]=i;
					t.setup(&pts[0]);
					triangles.push_back(t);
				}
			}
		} else { /* read mesh from files */
			unsigned int nnode=0; nf>>nnode; getline(nf,crap); /* first line: # nodes */
			for (i=0;i<nnode;i++) {
				int nodeno; float x,y;
				nf>>nodeno>>x>>y; std::getline(nf,crap); /* node lines */
				particle p;
				p.P=0.15*vec3(x,y,0)+vec3(0.5); 
				p.P.z=0.0001;
				p.V=vec3(0.0);
				if (p.P.x>0.5) p.V.y+=0.1; // initial rotation 
				//p.V+=0.01*(rv()-vec3(0.5)); // random jangling
				// +1.0*vec3(-p.P.y,p.P.x,0);  // rotation
				p.V.z=0.0;
				p.mass=1.0e-3; /* also added per triangle */
				p.C=rv();
				pts.push_back(p);
			}

			std::ifstream tf("mesh/image.1.ele"); 
			unsigned int ntri=0;
			tf>>ntri; std::getline(tf,crap);
			for (i=0;i<ntri;i++) {
				triangle t;
				int trino, a,b,c;
				tf>>trino>>a>>b>>c; std::getline(tf,crap);
				t.corner[0]=a-1;
				t.corner[1]=b-1;
				t.corner[2]=c-1;
				t.setup(&pts[0]);
				triangles.push_back(t);
			}
			
			for (int p=0;p<pts.size();p++) printf("Mass of %i: %f\n",p,pts[p].mass);
		}
		
		first_time=true;
	}
	
	/* Subcycle, so the physics runs at the right stepsize, 
	   yet still keeps up with the display. */
	double wall_time=oglTimer()*1.0; // <- for slow-motion 
	if (key_down[(int)' ']) sim_time=wall_time+1.0e-3; /* spacebar == stop physics */
	while (sim_time<wall_time)
	{ /* take one physics step */
		sim_time+=dt;
		
		float energy=0.0;
		/* zero net force */
		for (i=0;i<pts.size();i++) pts[i].F=vec2(0.0);
		
		/* Apply user interface force to the first particle */
		int keypart=pts.size()/2;
		if (key_down[(int)'x']) pts[keypart].F.x+=10.0;
		if (key_down[(int)'X']) pts[keypart].F.x-=10.0;
		if (key_down[(int)'y']) pts[keypart].F.y+=10.0;
		if (key_down[(int)'Y']) pts[keypart].F.y-=10.0;
		//if (key_down[(int)'z']) pts[keypart].F.z+=50.0;
		
		/* physics the triangles to find the net force on all points */
		for (i=0;i<triangles.size();i++) {
			triangle &t=triangles[i];
			t.physics(&pts[0]);
		}
		
		/* physics the particles */
		for (i=0;i<pts.size();i++) {
			particle &p=pts[i];

		/* Apply physics */
			vec3 gA(0.0,-0.1,0.0); /* gravity acceleration */

			vec3 A=vec3(p.F.x,p.F.y,0)/p.mass +gA; /* total acceleration */
			
			energy+=p.mass*dot(p.V,p.V);
			
			p.P+=dt*p.V+0.5*dt*dt*A; /* move position by velocity */
			p.V+=dt*A; /* accelerate particles downwards */
			
			/* Bounce off all six sides of cube */
			for (int dir=-1;dir<=1;dir+=2)
			for (int axis=0;axis<3;axis++) {
				float loc=(p.P[axis]-0.5)*2.0;
				if (loc*dir>1.0 && p.V[axis]*dir>0.0) { /* bounce off this axis */
					p.V[axis]*=-1.0;
				}
			}
		}
		
		if (energy>1.0e3) printf("t=%.3f, energy=%f\n",sim_time,energy);
		if (energy!=energy) exit(1); /* NaN energy: unstable */

		first_time=false;
	}
	
/* Main camera setup */
	glViewport(0,0,glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	glClearColor(0.3,0.5,0.7,1.0); /* clear to a nice blue color */
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(70.0, /* degree FOV */
		glutGet(GLUT_WINDOW_WIDTH)/float(glutGet(GLUT_WINDOW_HEIGHT)),
		0.001,1000.0 /* Z range */
	);
	oglCameraLookAt(1.0);
	
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// Draw in a ground plane
	glDisable(GL_DEPTH_TEST); /* don't read/write from depth buffer */
	//glDepthMask(GL_FALSE); /* don't write to depth buffer (no Z fighting!) */
	
	glBegin(GL_QUADS);
	glColor4f(0.6,0.4,0.2,1.0); /* light, opaque brown */
	glVertex3f(-big,-big,floor);
	glVertex3f(+big,-big,floor);
	glVertex3f(+big,+big,floor);
	glVertex3f(-big,+big,floor);
	glEnd();

	// Draw in a 1-meter grid on the ground
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_BLEND);
	glLineWidth(0.4);
	glBegin(GL_LINES);
	glColor4f(0.0f,0.0f,0.0f,0.3f);
	for (float v=-10;v<=10;v+=0.1) {
		glVertex3f(-big,v,floor); /* X lines */
		glVertex3f(+big,v,floor);
		glVertex3f(v,-big,floor); /* Y lines */
		glVertex3f(v,+big,floor);
	}
	glEnd();
	
	glDepthMask(GL_TRUE); /* depth writes back on */
	
/* Draw axes */
	oglCameraAxes();

/* Draw triangles */
	glLineWidth(1.0);
	for (i=0;i<triangles.size();i++) {
		triangle &t=triangles[i];
		
		glColor3f(0,0,0); /* black outline */
		glBegin(GL_LINE_LOOP);
		for (int j=0;j<3;j++) glVertex2fv(pts[t.corner[j]].P);
		glEnd();
		
		if (i<10) /* draw stress on selected triangles */
			t.draw(&pts[0]);
	}
	
/* Draw our points */
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_POINT_SMOOTH);
	glPointSize(3.0);
	
	glBegin(GL_POINTS);
	for (i=0;i<pts.size();i++) {
		particle &p=pts[i];
	/* Draw the particle */
		glColor3fv(p.C);
		glVertex3fv(p.P);
	}
	glEnd();
	
	glutSwapBuffers(); /* display everything we've rendered so far */
	glutPostRedisplay(); /* continual animation */
	
	// Print frames per second
	glFinish(); elapsed=oglTimer()-start_t; /* stop the stopwatch */
	static double last_print=0.0; /* Periodically print out the time */
	if (start_t>last_print+0.1) {
		display_dt=elapsed;
		printf("%.1f frames/sec, %.2f ms/frame\n",
			1.0/elapsed,1000.0*elapsed);
		last_print=start_t;
	}
	oglCheck("after display function");
}

int main(int argc,char **argv) {
	camera=vec3(0.5,0.5,1.0);
	camera_orient.y=vec3(0,1,0); camera_orient.z=vec3(0,0,1);

	glutInit(&argc,argv);  /* set up GLUT */
	glutInitWindowPosition(0,60);
	glutInitWindowSize(1000,700);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH);
	glutCreateWindow("FEM demo (try WASD and XYZ keys!)"); /* make a window */
	glewInit(); /* only AFTER create window! */
	glutDisplayFunc(display); /* here's how to draw the window */
	oglCameraInit();
	glutMainLoop(); /* start drawing */
}
